"""Fileman gRPC services."""
from pathlib import Path
from uuid import UUID

from fred_api.fileman.service_fileman_grpc_pb2 import ReadReply, ReadRequest, StatReply, StatRequest
from fred_api.fileman.service_fileman_grpc_pb2_grpc import FileServicer as BaseFileServicer
from grpc import StatusCode
from grpc._server import _Context
from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.orm.exc import NoResultFound

from .models import File


class FileServicer(BaseFileServicer):
    """File gRPC service.

    Attributes:
        storage: A path to the root of the storage.
        session_cls: SQLAlchemy Session class to be used for queries.
    """

    def __init__(self, storage: str, session_cls: Session = None):
        self.storage = Path(storage)
        self.session_cls = session_cls or sessionmaker()

    def stat(self, request: StatRequest, context: _Context) -> StatReply:
        """Return file metadata."""
        try:
            uuid = UUID(request.uuid.value)
        except ValueError:
            context.abort(StatusCode.INVALID_ARGUMENT, "'{}' is not a valid UUID".format(request.uuid.value))

        session = self.session_cls()
        try:
            file = session.query(File).filter_by(uuid=uuid).one()
        except NoResultFound:
            context.abort(StatusCode.NOT_FOUND, "File {} not found".format(uuid))
        except DBAPIError as error:
            context.abort(StatusCode.FAILED_PRECONDITION, "Database error: {}".format(error))
        finally:
            session.close()
        reply = StatReply()
        reply.data.create_datetime.FromDatetime(file.create_datetime)
        reply.data.size = file.size
        reply.data.mimetype = file.mimetype
        reply.data.name = file.name
        return reply

    def read(self, request: ReadRequest, context: _Context) -> ReadReply:
        """Return file content."""
        try:
            uuid = UUID(request.uuid.value)
        except ValueError:
            context.abort(StatusCode.INVALID_ARGUMENT, "'{}' is not a valid UUID".format(request.uuid.value))

        session = self.session_cls()
        try:
            file_path = session.query(File.path).filter_by(uuid=uuid).one()
        except NoResultFound:
            context.abort(StatusCode.NOT_FOUND, "File {} not found".format(uuid))
        except DBAPIError as error:
            context.abort(StatusCode.FAILED_PRECONDITION, "Database error: {}".format(error))
        finally:
            session.close()

        chunk_size = request.size or -1
        try:
            # XXX: Explicitly convert Path to str to support Python 3.5. `open` accepts Path since 3.6.
            with open(str(self.storage.joinpath(file_path[0])), mode='rb') as reader:
                while True:
                    reply = ReadReply()
                    chunk = reader.read(chunk_size)
                    reply.data.data = chunk
                    yield reply

                    if chunk_size == -1 or len(chunk) < chunk_size:
                        # If the file is read, terminate the generator.
                        return
        except OSError as error:
            context.abort(StatusCode.FAILED_PRECONDITION, "File {} is not available: {}".format(uuid, error))
