"""Fileman settings management."""
import os
from typing import Any, Dict, Iterable, Tuple, cast

from easysettings import YAMLSettings

ENVIRON_NAME = 'FILEMAN_CONFIG'
CONFIG_FILES = ('~/.fred/fileman.conf', '/etc/fred/fileman.conf')
DEFAULT_SETTINGS = {
    'db_connection': 'sqlite:///:memory:',
    'db_echo': False,
    'db_poolclass': None,
    'grpc_port': 50051,
    'storage_root': '/var/lib/fileman',
}
# Mapping command line options to setting names.
OPTIONS_MAPPING = {
    '--connection': 'db_connection',
    '--poolclass': 'db_poolclass',
    '--port': 'grpc_port',
    '--storage': 'storage_root',
}


class Settings:
    """Settings manager."""

    def __init__(self):
        self._settings = None

    def load(self, config_argv: str = None) -> None:
        """Return setting object from the first config file found.

        Raises:
            FileNotFoundError: If the config file is provided, but the file doesn't exist.
            RuntimeError: If settings are already loaded.
        """
        if self._settings is not None:
            raise RuntimeError('Settings already loaded')
        settings = YAMLSettings(DEFAULT_SETTINGS)
        if ENVIRON_NAME in os.environ:
            settings.load(os.environ[ENVIRON_NAME])
        elif config_argv is not None:
            settings.load(config_argv)
        else:
            for file in CONFIG_FILES:
                try:
                    settings.load(os.path.expanduser(file))
                    # Break the cycle if settings are loaded correctly.
                    break
                except FileNotFoundError:
                    continue
        self._settings = settings

    def __getitem__(self, key: str) -> Any:
        if self._settings is None:
            self.load()
        return cast(YAMLSettings, self._settings)[key]

    def __setitem__(self, key: str, value: Any) -> None:
        if self._settings is None:
            self.load()
        cast(YAMLSettings, self._settings)[key] = value

    def items(self) -> Iterable[Tuple[str, Any]]:
        """Return an iterator of pairs of all settings."""
        if self._settings is None:
            self.load()
        return cast(YAMLSettings, self._settings).items()

    def set_options(self, options: Dict[str, Any]):
        """Replace configuration values from command line options.

        Arguments:
            options: Command line argument parsed by docopt.
        """
        for option, setting in OPTIONS_MAPPING.items():
            value = options.get(option)
            if value:
                self[setting] = value


class SettingsProxy:
    """Proxy for settings instance.

    Attributes:
        settings: Actual settings
    """

    def __init__(self, settings: Settings):
        self.settings = settings

    def __getattr__(self, name: str) -> Any:
        return getattr(self.settings, name)

    def __getitem__(self, key: str) -> Any:
        return self.settings[key]

    def __setitem__(self, key: str, value: Any) -> Any:
        self.settings[key] = value


SETTINGS = SettingsProxy(Settings())
